a {
  filter: initial;
}

b {
  filter: inherit;
}

c {
  filter: unset;
}

d {
  filter: none;
}

e {
  filter: blur(2px) brightness(20%) contrast(120%) grayscale(100%) hue-rotate(45deg) invert(50%) opacity(25%) saturate(90%) sepia(100%);
}

f {
  filter: drop-shadow(1px 2px rgb(255,0,0));
}
