/*
 * Copyright © 2019 Benjamin Otte
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library. If not, see <http://www.gnu.org/licenses/>.
 *
 * Authors: Benjamin Otte <otte@gnome.org>
 */

#include "config.h"

#include "gtkcsserror.h"

/**
 * gtk_css_parser_error_quark:
 *
 * Registers an error quark for CSS parsing errors.
 *
 * Returns: the error quark
 **/
GQuark
gtk_css_parser_error_quark (void)
{
  return g_quark_from_static_string ("gtk-css-parser-error-quark");
}

/**
 * gtk_css_parser_warning_quark:
 *
 * Registers an error quark for CSS parsing warnings.
 *
 * Returns: the warning quark
 **/
GQuark
gtk_css_parser_warning_quark (void)
{
  return g_quark_from_static_string ("gtk-css-parser-warning-quark");
}

