@define-color some-color currentcolor;

a {
  color: currentcolor;
}

b {
  color: shade(currentcolor, 1.5);
}

c {
  color: alpha(currentcolor, 0.5);
}

d {
  color: mix(currentcolor, @some-color, 0.5);
}

e {
  color: mix(@some-color, currentcolor, 0.5);
}

f {
  color: color-mix(in srgb, rgb(255,0,0) 40%, currentcolor 60%);
}

g {
  color: color(from currentcolor srgb r g b / (alpha * 0.5));
}
