@keyframes aaaa {
  from {
    --test: red;
    --test2: green;
  }
  to {
    --test: blue;
    --test3: cyan;
  }
}

label {
  background-color: var(--color2, green);
  color: var(--color1);
  --extra-color: mix(var(--color1,), var(--color2, var(--color1)), var(--factor, 0.2));
}

:root {
  --color1: red;
  --color2: initial;
}
