a {
  font-feature-settings: initial;
}

b {
  font-feature-settings: inherit;
}

c {
  font-feature-settings: unset;
}

e {
  font-feature-settings: normal;
}

f {
  font-feature-settings: "dlig", "c2sc" off, "swsh" 2;
}
