/* DO NOT EDIT THIS FILE - it is machine generated */
#include <jni.h>
/* Header for class org_gtk_android_GlueLibraryContext */

#ifndef _Included_org_gtk_android_GlueLibraryContext
#define _Included_org_gtk_android_GlueLibraryContext
#ifdef __cplusplus
extern "C" {
#endif
/*
 * Class:     org_gtk_android_GlueLibraryContext
 * Method:    runApplication
 * Signature: (Landroid/app/Activity;)V
 */
JNIEXPORT void JNICALL Java_org_gtk_android_GlueLibraryContext_runApplication
  (JNIEnv *, jclass, jobject);

#ifdef __cplusplus
}
#endif
#endif
