* {
  all: unset;
}

window {
  background: pink;
}

grid {
  background-image: linear-gradient(rebeccapurple, rebeccapurple);
  background-size: 10px 10px;
  padding: 10px;
  background-repeat: repeat;
  background-origin: content-box;
  background-clip: border-box;
}

/* reference */
box {
  background: rebeccapurple;
}
